from django import template
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag(takes_context=True)
def attrs(context):
    """Convert context into a string of html attributes, converting '_' to '-'"""
    result = ""
    for key, value in context.flatten().items():
        if key not in ["True", "False", "None", "content", "element"]:
            if "hx_" in key:
                key = key.replace("_", "-")
            result += f' {key}="{value}"'
    return mark_safe(result)


@register.filter
def render_button(button):
    return button.render()


@register.filter
def td_attr(column, table):
    """
    Add td-edit class to td attributes if it is editable
    NB col.attrs is an immutable property
    """
    html = column.attrs["td"].as_html()
    td_edit = "td-edit"
    if column.name in table.columns_editable and td_edit not in html:
        value = column.attrs["td"]["class"]
        if value:
            s = html.replace(value, f"{td_edit} {value}")
        else:
            s = html + f' class="{td_edit}"'
        return mark_safe(s)
    return html


@register.filter
def has_filter_toolbar(view):
    return view.filterset is not None and view.filterstyle == view.filterstyle.TOOLBAR



from src.django_tableaux.views import TableauxView
from .models import Model1
from django_tables2 import Table
from django.shortcuts import render
from django_htmx.http import retarget
from src.django_tableaux.buttons import Button
from src.django_tableaux.columns import SelectionColumn


class View1(TableauxView):
    model = Model1


class Table1(Table):
    class Meta:
        model = Model1
        fields = (
            "name",
            "description",
            "decimal",
        )
        sequence = ["name", "select", "..."]
        attrs = {
            "class": "table",
        }

    select = SelectionColumn()


class View2(TableauxView):
    model = Model1
    table_class = Table1
    row_settings = True
    column_settings = True

    def get_queryset(self):
        if Model1.objects.count() < 10:
            create_objects(10)
        return Model1.objects.all()

    def get_bulk_actions(self):
        return (
            ("action_message", "Action with message"),
            ("export", "Export to csv"),
            ("export_xlsx", "Export as xlsx"),
        )

    def get_buttons(self):
        return [Button("Button 1"), Button("Button 2")]

    def handle_action(self, request, action):
        if action == "action_message":
            context = {
                "message": f"Action on {self.selected_objects.count()} rows",
                "alert_class": "alert-success",
            }
            response = render(request, "django_tableaux/_alert.html", context)
            return retarget(response, "#messages")


def create_objects(count):
    for x in range(count):
        Model1.objects.create(
            name=f"name_{x}", description=f"description_{x}", decimal=x
        )
    return Model1.objects.all()



from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "just-an-insecure-test-key"
DEBUG = True

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.humanize",
    "bootstrap4",
    "django_tables2",
    "django_filters",
    "django_htmx",
    "src.django_tableaux",
    "myapp",
    "tests",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django_htmx.middleware.HtmxMiddleware",
]

ROOT_URLCONF = "myapp.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Django_bootstrap4 settings
BOOTSTRAP4 = {"error_css_class": "", "success_css_class": ""}
DJANGO_TABLEAUX_LIBRARY = "bootstrap4"



# Generated by Django 4.2.3 on 2023-09-11 11:24

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Model1",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                ("description", models.CharField(max_length=100)),
                ("decimal", models.DecimalField(decimal_places=2, max_digits=10)),
            ],
        ),
    ]



import pytest
from django.db.models import *
from django.test import RequestFactory, TestCase
from django_tables2 import tables
from myapp.models import *
from django_htmx.middleware import HtmxDetails

from src.django_tableaux.views import TableauxView


class Table1(tables.Table):
    class Meta:
        model = Model1


class View1(TableauxView):
    model = Model1

    def get(self, request, *args, **kwargs):
        return super().get(request, *args, **kwargs)


# class MyView(TestCase):
#     def test_get(self):
#         request = RequestFactory().get("/")
#         view = TableauxView()
#         view.setup(request)


@pytest.mark.django_db
def test_empty_table():
    request = RequestFactory().get("/")
    request.htmx = ""
    request.session = {}
    response = View1.as_view()(request)
    assert response.status_code == 200
    response.render()
    assert "<table" in response.rendered_content
    assert "<td" not in response.rendered_content


@pytest.mark.django_db
def test_table_with_data():
    for x in range(2):
        Model1.objects.create(
            name=f"name_{x}", description=f"description_{x}", decimal=x
        )
    request = RequestFactory().get("/")
    request.htmx = ""
    request.session = {}
    response = View1.as_view()(request)
    assert response.status_code == 200
    response.render()
    p = response.rendered_content.find("<table")
    body = response.rendered_content[p:]
    assert body.count("<tr") == 10



from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from .utils import get_template_prefix


class Button:
    template_name = "button.html"

    def __init__(self, content, name="", typ="button", css="btn btn-primary", **kwargs):
        self.context = {
            "element": "a" if kwargs.get("href") else "button",
            "content": content,
            "name": slugify(content) if not name else name,
            "class": css,
        }
        if self.context["element"] == "button":
            self.context["type"] = typ
        self.context.update(kwargs)

    def render(self):
        html = mark_safe(
            render_to_string(
                template_name=f"{get_template_prefix()}/{self.template_name}",
                context=self.context,
            )
        )
        return html



from django.db.models import *
from django_tables2 import tables

from src.django_tableaux.buttons import Button
from src.django_tableaux.columns import SelectionColumn
from src.django_tableaux.utils import define_columns, set_select_column
from myapp.models import TestModel


class TestTable1(tables.Table):
    class Meta:
        model = TestModel


def test_define_columns_no_fields_means_all_fixed():
    table = TestTable1([])
    define_columns(table, width=0)
    assert table.columns_fixed == ["id"]
    assert table.columns_optional == ["a", "b", "c", "d"]
    assert table.columns_default == ["id", "a", "b", "c", "d"]


class TestTable2(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c"]


def test_define_columns_fields():
    table = TestTable2([])
    define_columns(table, width=0)
    assert table.columns_fixed == ["a"]
    assert table.columns_optional == ["b", "c"]
    assert table.columns_default == ["a", "b", "c"]


class TestTable3(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c"]
        sequence = ["c", "b", "a"]


def test_define_columns_sequence():
    table = TestTable3([])
    define_columns(table, width=0)
    assert table.columns_fixed == ["c"]
    assert table.columns_optional == ["b", "a"]
    assert table.columns_default == ["c", "b", "a"]


class TestTable4(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c"]
        columns = {"fixed": ["a", "b"], "default": ["a", "b", "c"]}


def test_define_columns_fixed_and_default():
    table = TestTable4([])
    define_columns(table, width=0)
    assert table.columns_fixed == ["a", "b"]
    assert table.columns_optional == ["c"]
    assert table.columns_default == ["a", "b", "c"]


class TestTable5(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c", "d"]
        columns = {"fixed": ["a", "b", "c", "d"]}
        responsive = {
            100: {"fixed": ["a"]},
            500: {"fixed": ["a", "b"], "default": ["a", "b", "c"]},
            1000: {"fixed": ["a", "b", "c"], "default": ["a", "b", "c", "d"]},
        }


def test_define_columns_responsive():
    table = TestTable5([])
    define_columns(table, width=100)
    assert table.columns_fixed == ["a"]
    assert table.columns_optional == ["b", "c", "d"]
    assert table.columns_default == ["a", "b", "c", "d"]
    define_columns(table, width=500)
    assert table.columns_fixed == ["a", "b"]
    assert table.columns_optional == ["c", "d"]
    assert table.columns_default == ["a", "b", "c"]
    define_columns(table, width=1000)
    assert table.columns_fixed == ["a", "b", "c"]
    assert table.columns_optional == ["d"]
    assert table.columns_default == ["a", "b", "c", "d"]
    assert table.columns_default == ["a", "b", "c", "d"]


class TestTable6(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c", "d"]
        responsive = {
            100: {"fixed": ["a"]},
            1000: {},
        }


def test_define_columns_responsive_no_defaults():
    table = TestTable6([])
    define_columns(table, width=100)
    assert table.columns_fixed == ["a"]
    ## default=all fields if default not specified
    assert table.columns_default == ["a", "b", "c", "d"]
    define_columns(table, width=1000)
    # Only first field is fixed if fixed not specified
    assert table.columns_fixed == ["a"]
    # default is every field if no fixed and no default
    assert table.columns_default == ["a", "b", "c", "d"]


class TestTable7(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c", "d"]

    selection = SelectionColumn()


def test_define_columns_with_selection_not_in_sequence():
    table = TestTable7([])
    set_select_column(table)
    define_columns(table, width=0)
    assert table.columns_fixed == ["selection", "a"]


class TestTable8(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c", "d"]
        sequence = ["selection", "..."]

    selection = SelectionColumn()


def test_define_columns_with_selection_first_in_sequence():
    table = TestTable8([])
    table.select_name = "selection"
    define_columns(table, width=0)
    assert table.columns_fixed == ["selection", "a"]


class TestTable9(tables.Table):
    class Meta:
        model = TestModel
        fields = ["a", "b", "c", "d"]
        sequence = ["a", "selection", "..."]

    selection = SelectionColumn()


def test_define_columns_with_selection_not_first_in_sequence():
    table = TestTable9([])
    table.select_name = "selection"
    define_columns(table, width=0)
    assert table.columns_fixed == ["a", "selection"]


def test_default_button():
    button = Button("Test button")
    html = button.render()
    assert "<button" in html
    assert 'type="button"' in html
    assert 'class="btn btn-primary"' in html
    assert 'name="test-button"' in html
    assert ">Test button</button" in html


def test_button_renders_attributes():
    button = Button(
        "Test button",
        css="btn btn-secondary",
        type="submit",
        name="test_name",
        hx_get="/test",
        hx_target="#target",
    )
    html = button.render()
    assert 'class="btn btn-secondary"' in html
    assert 'type="submit"' in html
    assert 'name="test_name"' in html
    assert 'hx-get="/test"' in html
    assert 'hx-target="#target"' in html


def test_link_button_when_href_present():
    button = Button("Test button", href="url")
    html = button.render()
    assert "<a" in html
    assert 'href="url"' in html
    assert "type" not in html


# def test_update_url_empty():
#     url = "http://test.com/"
#     result = update_url(url, "per_page", 25)
#     assert result == "http://test.com/?per_page=25"
#
#
# def test_update_url_existing_empty():
#     url = "http://test.com/?per_page="
#     result = update_url(url, "per_page", 25)
#     assert result == "http://test.com/?per_page=25"
#
#
# def test_update_url_existing_empty_middle():
#     url = "http://test.com/?x=1&per_page=&y=2"
#     result = update_url(url, "per_page", 25)
#     assert result == "http://test.com/?x=1&per_page=25&y=2"
#
#
# def test_update_url_existing_empty_end():
#     url = "http://test.com/?x=1&per_page="
#     result = update_url(url, "per_page", 25)
#     assert result == "http://test.com/?x=1&per_page=25"
#
#
# def test_update_url_existing_sole():
#     url = "http://test.com/?per_page=10"
#     result = update_url(url, "per_page", 25)
#     assert result == "http://test.com/?per_page=25"
#
#
# def test_update_url_existing_middle():
#     url = "http://test.com/?x=1&per_page=10&y=2"
#     result = update_url(url, "per_page", 25)
#     assert result == "http://test.com/?x=1&per_page=25&y=2"
#
#
# def test_update_url_quoted_string():
#     url = "http://test.com/?x=1&per_page=10&y=2"
#     result = update_url(url, "per_page", "a&b=c")
#     assert result == "http://test.com/?x=1&per_page=a%26b%3Dc&y=2"



"""
Django tables with Advanced User eXperience
===============
This project builds upon two well-established django apps: `django_tables2 <https://github.com/jieter/django-tables2>`_
and `django_filter <https://github.com/carltongibson/django-filter>`_ and enhances their functionality through a
sprinkling of `htmx <https://htmx.org>`_ magic. It provides a single class-based view in which you can enable multiple
features to deliver a customised user experience that embodies the best practice for interactive tables.

Key features
============
* The columns to display can be user-defined
* Fully responsive: Define different column layouts for mobile, tablet and desktop
* Bulk actions on selected rows or on all (possibly filtered) rows
* Infinite scroll
* Infinite load-more data
* Position filters in a toolbar, in a modal or embed them within the table header
* Show active filters with ability to clear individual filters
* Edit specific fields directly inside the table
* Easy integration with generic views for CRUD operations

The project is still in beta but is fully usable.

Full documentation to follow.
"""

__version__ = "0.6.8"



import pytest
import os
from src.django_tableaux.utils import define_columns, build_templates_dictionary


def test_default_templates():
    templates = build_templates_dictionary()
    assert "bootstrap4" in templates["render_body"]
    assert "bootstrap4" in templates["modal_base"]


def test_bootstrap5_templates():
    templates = build_templates_dictionary("bootstrap5")
    assert "bootstrap4" in templates["render_body"]
    assert "bootstrap5" in templates["modal_base"]


def test_external_templates():
    my_path = os.path.dirname(os.path.realpath(__file__))
    library = my_path + "/template_library"
    templates = build_templates_dictionary(library)
    assert "template_library" in templates["render_body"]
    assert "template_library" in templates["modal_base"]



from django.db import models


class Model1(models.Model):
    name = models.CharField(max_length=50)
    description = models.CharField(max_length=100)
    decimal = models.DecimalField(max_digits=10, decimal_places=2)


class TestModel(models.Model):
    a = models.CharField(max_length=20)
    b = models.CharField(max_length=20)
    c = models.CharField(max_length=20)
    d = models.CharField(max_length=20)



import itertools
import django_tables2 as tables
from django.contrib.humanize.templatetags.humanize import intcomma
from .utils import get_template_prefix
from django.utils.safestring import mark_safe


class RightAlignedColumn(tables.Column):
    def __init__(self, **kwargs):
        super().__init__()
        if "th" not in self.attrs:
            self.attrs["th"] = {}
        if "td" not in self.attrs:
            self.attrs["td"] = {}
        self.attrs["th"]["style"] = "text-align: right;"
        self.attrs["td"]["style"] = "text-align: right;"


class CenteredColumn(tables.Column):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if "th" not in self.attrs:
            self.attrs["th"] = {}
        if "td" not in self.attrs:
            self.attrs["td"] = {}
        self.attrs["th"]["style"] = "text-align: center;"
        self.attrs["td"]["style"] = "text-align: center;"


class CenteredTrueColumn(CenteredColumn):
    def render(self, value):
        if value:
            return "\u2705"
        return ""


class CenteredTrueFalseColumn(CenteredColumn):
    def render(self, value):
        if value:
            return "\u2705"
        return "\u274c"


class CurrencyColumn(RightAlignedColumn):
    def __init__(self, **kwargs):
        self.integer = kwargs.pop("integer", None)
        self.prefix = kwargs.pop("prefix", "")
        self.suffix = kwargs.pop("suffix", "")
        super().__init__(**kwargs)

    def render(self, value):
        if self.integer:
            value = int(value)
        return mark_safe(f"{self.prefix}{intcomma(value)}{self.suffix}")


class CheckBoxColumn(tables.TemplateColumn):
    def __init__(self, **kwargs):
        kwargs["template_name"] = f"{get_template_prefix()}/custom_checkbox.html"
        super().__init__(**kwargs)


class SelectionColumn(tables.TemplateColumn):
    def __init__(self, **kwargs):
        kwargs["template_name"] = f"{get_template_prefix()}/select_checkbox.html"
        kwargs["verbose_name"] = ""
        kwargs["accessor"] = "id"
        kwargs["orderable"] = False
        super().__init__(**kwargs)


class CounterColumn(tables.Column):
    row_counter = None

    def __init__(self, **kwargs):
        kwargs["orderable"] = False
        kwargs["empty_values"] = ()
        kwargs["verbose_name"] = ""
        super().__init__(**kwargs)

    def render(self, value):
        if not self.row_counter:
            self.row_counter = itertools.count()
        return next(self.row_counter) + 1



import logging
from enum import IntEnum
from os import listdir
from os.path import isfile, join

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import QueryDict, HttpResponse
from django.shortcuts import render, reverse
from django.template.response import TemplateResponse
from django.urls.resolvers import NoReverseMatch
from django.views.generic import TemplateView
from django_filters.filterset import filterset_factory
from django_htmx.http import (
    HttpResponseClientRefresh,
    HttpResponseClientRedirect,
    trigger_client_event,
)
from django_tables2 import SingleTableMixin
from django_tables2.export.export import TableExport

from .utils import (
    breakpoints,
    define_columns,
    set_select_column,
    set_column_states,
    save_columns,
    load_columns,
    set_column,
    visible_columns,
    save_per_page,
    build_templates_dictionary,
)

logger = logging.getLogger(__name__)


class TableauxView(SingleTableMixin, TemplateView):
    class FilterStyle(IntEnum):
        NONE = 0
        TOOLBAR = 1
        MODAL = 2
        HEADER = 3

    class ClickAction(IntEnum):
        NONE = 0
        GET = 1
        HX_GET = 2
        CUSTOM = 3

    title = ""
    template_name = "django_tableaux/bootstrap4/tableaux.html"
    template_library = "bootstrap4"

    model = None
    form_class = None
    filterset_class = None
    filterset_fields = None
    filterset = None

    table_pagination = {"per_page": 10}
    infinite_scroll = False
    infinite_load = False
    #
    context_filter_name = "filter"
    filter_style = FilterStyle.NONE
    filter_pills = False
    filter_button = False  # only relevant for TOOLBAR style
    #
    column_settings = False
    row_settings = False

    click_action = ClickAction.NONE
    click_url_name = ""
    click_target = "#modals-here"
    #
    sticky_header = False
    buttons = []
    object_name = ""
    #
    export_filename = "table"
    export_format = "csv"
    export_class = TableExport
    export_formats = (TableExport.CSV,)

    ALLOWED_PARAMS = ["page", "per_page", "sort", "_width"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.table = None
        self.width = 0
        self.object_list = None
        self.selected_objects = None
        self.selected_ids = None
        self.templates = build_templates_dictionary()

    def get(self, request, *args, **kwargs):
        table_class = self.get_table_class()
        # If table is responsive and no width parameter was sent,
        # tell client to repeat request adding the width parameter
        if hasattr(table_class, "Meta") and hasattr(table_class.Meta, "responsive"):
            if "_width" in request.GET:
                self.width = int(request.GET["_width"])
            else:
                return render(request, self.templates["width_request"])

        if request.htmx:
            return self.get_htmx(request, *args, **kwargs)

        if "_export" in request.GET:
            return self.export_table()

        return self.render_table()

    def get_queryset(self):
        if hasattr(self, "queryset"):
            return self.queryset
        if self.model is not None:
            return self.model._default_manager.all()
        else:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. Define "
                "%(cls)s.model, %(cls)s.queryset, or override "
                "%(cls)s.get_queryset()." % {"cls": self.__class__.__name__}
            )

    def render_table(self, template_name=None):
        """
        Use the TemplateResponse Mixin to render the table, optionally using a specific template
        """
        self.object_list = self.get_queryset()
        self.filterset = self.get_filterset(self.object_list)
        if self.filterset is not None:
            self.object_list = self.filterset.qs
        self.table = self.get_table()
        self.preprocess_table(self.table, self.filterset)
        context = self.get_context_data()
        template_name = template_name or self.template_name
        return self.render_to_response(template_name, context)

    def render_to_response(self, template_name, context, **response_kwargs):
        response_kwargs.setdefault("content_type", self.content_type)
        return TemplateResponse(
            request=self.request,
            template=[template_name],
            context=context,
            **response_kwargs,
        )

    def export_table(self):
        self.object_list = self.get_queryset()
        subset = self.request.GET.get("_subset", None)
        if subset:
            if subset == "selected":
                self.object_list = self.object_list.filter(
                    id__in=self.request.session.get("selected_ids", [])
                )
            elif subset == "all":
                self.filterset = self.get_filterset(self.object_list)
                self.object_list = self.filterset.qs
        export_format = self.request.GET.get("_export", self.export_format)
        # Use tablib to export in desired format
        table = self.get_table()
        self.preprocess_table(table)
        exclude_columns = [k for k, v in table.columns.columns.items() if not v.visible]
        exclude_columns.append("selection")
        exporter = self.export_class(
            export_format=export_format,
            table=table,
            exclude_columns=exclude_columns,
        )
        return exporter.response(filename=f"{self.export_filename}.{export_format}")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            table=self.table,
            filter=self.filterset,
            object_list=self.object_list,
            title=self.title,
            filter_button=self.filter_button,
            filter_pills=self.filter_pills,
            filters=[],
            buttons=self.get_buttons(),
            actions=self.get_bulk_actions(),
            rows=self.rows_list(),
            per_page=self.request.GET.get(
                "per_page", self.table_pagination.get("per_page", 25)
            ),
            breakpoints=breakpoints(self.table),
            width=self.width,
            templates=self.templates,
        )
        if "_width" in self.request.GET:
            context["breakpoints"] = None

        for key, value in self.request.GET.items():
            if key not in self.ALLOWED_PARAMS and value:
                context["filters"].append((key, value))
        return context

    def get_filterset(self, queryset=None):
        filterset_class = self.filterset_class
        filterset_fields = self.filterset_fields

        if filterset_class is None and filterset_fields:
            filterset_class = filterset_factory(self.model, fields=filterset_fields)

        if filterset_class is None:
            return None

        return filterset_class(
            self.request.GET,
            queryset=queryset,
            request=self.request,
        )

    def get_htmx(self, request, *args, **kwargs):
        # Some actions depend on trigger_name; others on trigger
        trigger_name = request.htmx.trigger_name
        trigger = request.htmx.trigger
        if trigger_name:
            if trigger_name == "filter" and self.filterset_class:
                # show filter modal
                context = {"filter": self.filterset_class(request.GET)}
                return render(request, self.templates["modal_filter"], context)

            elif trigger_name == "filter_form":
                # a filter value was changed
                return self.render_table(self.templates["render_table_data"])

            elif "clr_" in trigger_name:
                # cancel a filter
                filter = trigger_name.split("_")[1]
                qd = request.GET.copy()
                if filter == "all":
                    keys = list(qd.keys())
                    for key in keys:
                        if key not in self.ALLOWED_PARAMS:
                            qd.pop(key)
                else:
                    qd.pop(filter)
                return HttpResponseClientRedirect(f"{request.path}?{qd.urlencode()}")

        if trigger == "table_data":
            # triggered refresh of table data after create or update
            return self.render_table(self.templates["render_body"])

        elif "id_row" in trigger:
            # change number of rows to display
            rows = trigger_name
            save_per_page(request, rows)
            url = self._update_parameter(request, "per_page", rows)
            return HttpResponseClientRedirect(url)

        elif "tr_" in trigger:
            # infinite scroll/load_more or click on row
            if "_scroll" in request.GET:
                return self.render_table(self.templates["render_body"])

            return self.row_clicked(
                pk=trigger.split("_")[1],
                target=request.htmx.target,
                return_url=request.htmx.current_url,
            )

        elif "cell_" in trigger:
            # cell clicked
            bits = trigger.split("_")
            return self.edit_cell(
                pk=bits[1],
                column_name=visible_columns(request, self.table_class, self.width)[
                    int(bits[2])
                ],
                target=request.htmx.target,
            )

        elif "td_" in trigger:
            # cell clicked
            bits = trigger.split("_")
            return self.cell_clicked(
                pk=bits[1],
                column_name=visible_columns(request, self.table_class, self.width)[
                    int(bits[2])
                ],
                target=request.htmx.target,
            )

        # Column handling
        elif "id_col_reset" in trigger:
            # Reset default columns settings.
            # To make sure the column drop down is correctly updated we do a full client refresh,
            table = self.table_class([])
            define_columns(table, self.width)
            save_columns(request, self.width, table.columns_default)
            return HttpResponseClientRefresh()

        elif "id_col" in trigger:
            # Click on a column checkbox in the dropdown re-renders the table data with new column settings.
            # The column dropdown does not need to be rendered because the checkboxes are in the correct state,
            col_name = trigger_name[5:]
            checked = trigger_name in request.GET
            set_column(request, self.width, col_name, checked)
            return self.render_table(self.templates["render_table_data"])

        elif "id_" in trigger:
            if trigger == request.htmx.target:
                # Clear filter value triggered by  click on X in input-prepend
                qd = request.GET.copy()
                qd.pop(trigger_name)
                return HttpResponseClientRedirect(f"{request.path}?{qd.urlencode()}")
            #
            # Filter value changed
            url = self._update_parameter(
                request, trigger_name, request.GET.get(trigger_name, "")
            )
            return HttpResponseClientRedirect(url)
        raise ValueError("Bad htmx get request")

    def rows_list(self):
        return [10, 15, 20, 25, 50, 100]

    def get_buttons(self):
        return []

    def get_bulk_actions(self):
        """
        Return a list or tuple with element in format (action_name, "Action text")
        """
        return []

    def patch(self, request, *args, **kwargs):
        # PATCH is used to update a cell after inline editing
        params = QueryDict(request.body)
        bits = request.htmx.target.split("_")
        column_name = visible_columns(request, self.table_class, int(bits[3]))[
            int(bits[2])
        ]
        value = params.get(column_name, None)
        print(bits[1], value)
        if value:
            return self.cell_changed(
                record_pk=bits[1],
                column_name=column_name,
                value=params[column_name],
                target=request.htmx.target,
            )
        return HttpResponse("x")

    def post(self, request, *args, **kwargs):
        # Posts are an action performed on a queryset
        if "select_all" in request.POST:
            subset = "all"
            self.selected_ids = []
            self.selected_objects = self.filtered_query_set(request)
        else:
            subset = "selected"
            self.selected_ids = request.POST["selected_ids"].split(",")
            self.selected_objects = self.get_queryset().filter(pk__in=self.selected_ids)

        if request.htmx.trigger_name:
            if "export" in request.htmx.trigger_name:
                # Export is a special case. It redirects to a regular GET that returns the file
                request.session["selected_ids"] = self.selected_ids
                bits = request.htmx.trigger_name.split("_")
                export_format = bits[1] if len(bits) > 1 else "csv"
                path = request.path + request.POST["query"]
                if len(request.POST["query"]) > 1:
                    path += "&"
                return HttpResponseClientRedirect(
                    f"{path}_export={export_format}&_subset={subset}"
                )

            response = self.handle_action(request, request.htmx.trigger_name)
            if response:
                return response
        return HttpResponseClientRefresh()

    def get_filterset_kwargs(self, filterset_class):
        kwargs = super().get_filterset_kwargs(filterset_class)
        if kwargs["data"]:
            qd = kwargs["data"].copy()
            if "filter_save" in qd.keys():
                qd.pop("filter_save")
                kwargs["data"] = qd
        return kwargs

    def filtered_query_set(self, request, next=False):
        """Recreate the queryset used in GET for use in POST"""
        query_set = self.get_queryset()
        qd = self.query_dict(request)
        if next:
            if "page" not in qd:
                qd["page"] = "2"
            else:
                qd["page"] = str(int(qd["page"]) + 1)
        if self.filterset_class:
            return self.filterset_class(qd, queryset=query_set, request=request).qs
        return query_set

    def query_dict(self, request):
        bits = request.htmx.current_url.split("?")
        if len(bits) == 2:
            return QueryDict(bits[1]).copy()
        return QueryDict().copy()

    def handle_action(self, request, action):
        """
        self.selected_objects is a queryset that contains the objects to be processed.
        self.selected_ids is a list of model ids that were selected, empty for 'All rows'
        Possible return values:
        - None: (default) - reloads the last path
        - HttpResponse to be returned
        """
        return None

    def cell_clicked(self, pk, column_name, target, return_url):
        """
        User clicked on a cell with custom action
        """
        return HttpResponseClientRefresh()

    def edit_cell(self, pk, column_name, target):
        """
        User clicked on an editable cell
        """
        if not self.model:
            raise ImproperlyConfigured(
                "Model must be specified or cell_clicked must be overriden for editable cells",
            )
        if not self.form_class:
            raise ImproperlyConfigured(
                "You must specify the form_class for editable cells"
            )
        try:
            record = self.model.objects.get(pk=pk)
            form = self.form_class({column_name: getattr(record, column_name)})
            context = {"field": form[column_name], "target": target}
            return render(self.request, self.templates["cell_form"], context)
        except Exception as e:
            raise ValueError(f"Error {str(e)} while editing field {column_name}")

    def cell_changed(self, record_pk, column_name, value, target):
        """
        Editable cell value changed
        """
        try:
            record = self.model.objects.get(pk=record_pk)
            setattr(record, column_name, value)
            record.save()
        except ValueError:
            return render(
                self.request,
                self.templates["cell_error"],
                {"error": "Value error", "column": column_name, "target": target},
            )
        return HttpResponseClientRefresh()

    def preprocess_table(self, table, _filter=None):
        """
        Add extra attributes needed for rendering to the table
        """
        table.filter = _filter
        table.infinite_scroll = self.infinite_scroll
        table.infinite_load = self.infinite_load
        table.sticky_header = self.sticky_header
        # variables that control action when table is clicked
        table.click_action = self.click_action
        table.url = ""
        table.pk = False
        if self.click_url_name:
            # handle case when there is no PK passed (create)
            try:
                table.url = reverse(self.click_url_name)
            except NoReverseMatch:
                # Detail or update views have a pk
                try:
                    table.url = reverse(self.click_url_name, kwargs={"pk": 0})[:-2]
                    table.pk = True
                except NoReverseMatch:
                    raise (
                        ImproperlyConfigured(
                            f"Cannot resolve click_url_name: '{self.click_url_name}'"
                        )
                    )
        table.target = self.click_target

        set_select_column(table)
        if self.get_bulk_actions() and not table.select_name:
            raise ImproperlyConfigured(
                "Bulk actions require a selection column to be defined"
            )
        if table.select_name and not self.get_bulk_actions():
            raise ImproperlyConfigured("Selection column without bulk actions")

        # define possible columns depending upon the current width
        define_columns(table, width=self.width)

        # set visible columns according to saved setting
        table.columns_visible = load_columns(self.request, width=self.width)
        if not table.columns_visible:
            table.columns_visible = table.columns_default
            save_columns(
                self.request, width=self.width, column_list=table.columns_visible
            )
        else:
            # ensure all fixed columns are in the visible list in case table definitions have been changed
            for entry in table.columns_fixed:
                if entry not in table.columns_visible:
                    table.columns_visible.append(entry)

        set_column_states(table)

        if table.filter:
            table.filter.style = self.filter_style
            # If filter is in header, build list of filters in same sequence as columns
            if self.filter_style == self.FilterStyle.HEADER:
                table.header_fields = []
                for col in table.sequence:
                    if table.columns.columns[col].visible:
                        if col in table.filter.base_filters.keys():
                            table.header_fields.append(table.filter.form[col])
                        else:
                            table.header_fields.append(None)

    @staticmethod
    def _update_parameter(request, key, value):
        query_dict = request.GET.copy()
        query_dict[key] = value
        return f"{request.path}?{query_dict.urlencode()}"

    def has_filter_toolbar(self):
        return (
            self.filterset is not None
            and self.filter_style == TableauxView.FilterStyle.TOOLBAR
        )

    def has_filter_pills(self):
        return self.filter_pills


class SelectedMixin:
    """
    Use in views that are called to perform an action on selected objects.
    Selected objects can be obtained from a list of ids passed in the session,
    or from a query passed as GET parameters.
    Returns a queryset of the selected objects
    """

    model = None
    filterset_class = None

    def get_query_set(self):
        if self.model is None:
            raise ImproperlyConfigured("Model must be specified for SelectedMixin")
        ids = self.request.session.get("selected_ids", [])
        if ids:
            return self.model.objects.filter(id__in=ids)
        query_set = self.model.objects.all()
        if self.filterset_class:
            return self.filterset_class(
                self.request.GET, queryset=query_set, request=self.request
            ).qs
        return query_set


class ModalMixin:
    """Mixin to convert generic views to operate as modal views when called by hx-get"""

    title = ""

    def get_template_names(self):
        # handle case where same view can have different templates
        if self.request.htmx:
            if hasattr(self, "modal_template_name"):
                return [self.modal_template_name]
        if hasattr(self, "template_name"):
            return [self.template_name]
        raise ValueError("Template name missing")

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        url = self.request.resolver_match.route
        if "<int:pk>" in url and self.object:
            url = url.replace("<int:pk>", str(self.object.pk))
        context["modal_url"] = "/" + url
        return context

    def reload_table(self):
        response = HttpResponse("")
        return trigger_client_event(
            response, "reload", {"url": self.request.htmx.current_url_abs_path}
        )


def bulk_action_namer(texts: list):
    result = []
    for text in texts:
        name = text.lower().replace(" ", "_")
        result.append((name, text))



from urllib.parse import urlparse, parse_qs, quote
from os import listdir
from os.path import isfile, join, exists
from django.http import HttpRequest
from django.conf import settings
from django.apps import apps
from django.core.exceptions import ImproperlyConfigured


def _view_name(request):
    try:
        return request.resolver_match.view_name
    except AttributeError:
        return "test"


def save_columns(request: HttpRequest, width: int, column_list: list):
    key = f"columns:{_view_name(request)}:{width}"
    request.session[key] = column_list


def load_columns(request: HttpRequest, width: int):
    key = f"columns:{_view_name(request)}:{width}"
    return request.session[key] if key in request.session else None


def set_column(
    request: HttpRequest, width: int, column_name: str, checked: bool
) -> list:
    columns = load_columns(request, width)
    if checked:
        if column_name not in columns:
            columns.append(column_name)
    elif column_name in columns:
        columns.remove(column_name)
    save_columns(request, width, columns)
    return columns


def visible_columns(request: HttpRequest, table_class, width):
    """
    Return the list of visible column names in correct sequence
    """
    table = table_class(data=[])
    define_columns(table, width)
    if not table.responsive:
        width = 0
    columns = load_columns(request, width)
    return [col for col in table.sequence if col in columns]


def set_select_column(table):
    """
    Set table.select_name to name of the (first) Selection column if one exists
    and update sequence if necessary
    """
    table.select_name = ""
    for col in table.columns:
        if col.column.__class__.__name__ == "SelectionColumn":
            table.select_name = col.name
            break

    if table.select_name:
        # if user has defined the position of the selection column in Meta.sequence, respect it
        if (
            hasattr(table, "Meta")
            and hasattr(table.Meta, "sequence")
            and table.select_name in table.Meta.sequence
        ):
            return
        # else make sure that selection is the first column
        index = table.sequence.index(table.select_name)
        if index > 0:
            table.sequence.remove(table.select_name)
            table.sequence.insert(0, table.select_name)


def define_columns(table, width):
    """
    Add 4 lists of column names to table:
    columns_fixed = columns that are always visible
    columns_optional = columns that the user can choose to show
    columns_default = default visible columns
    columns_editable = columns that can be edited in situ
    """
    table.responsive = False
    table.columns_fixed = []
    table.columns_default = table.sequence
    table.columns_editable = []
    # Ensure table.select_name exists for ease of testing
    if not hasattr(table, "select_name"):
        table.select_name = ""
    if table.Meta:
        col_dict = {}
        if hasattr(table.Meta, "editable"):
            table.columns_editable = table.Meta.editable
        if hasattr(table.Meta, "columns"):
            if type(table.Meta.columns) == dict:
                col_dict = table.Meta.columns
            else:
                raise ValueError("Meta.columns must be a dictionary")
        if hasattr(table.Meta, "responsive"):
            table.responsive = True
            key = 0
            for breakpoint in table.Meta.responsive.keys():
                if width >= breakpoint:
                    key = breakpoint
            col_dict = table.Meta.responsive.get(key, {})
            # todo attrs
        table.columns_fixed = col_dict.get("fixed", [])
        table.columns_default = col_dict.get("default", table.sequence)
        table.mobile = col_dict.get("mobile", False)
    if not table.columns_fixed:
        table.columns_fixed = table.sequence[:1]
        if table.columns_fixed[0] == table.select_name and len(table.sequence) > 1:
            table.columns_fixed = table.sequence[:2]
        if table.select_name and table.select_name not in table.columns_fixed:
            table.columns_fixed.append(table.select_name)
    table.columns_optional = [
        c
        for c in table.sequence
        if c not in table.columns_fixed and c != table.select_name
    ]


def set_column_states(table):
    """
    Control column visibility and
    add attribute 'columns_states' - a list of tuples used to create the column dropdown
    Expects 'table.columns_visible' to have been updated beforehand
    """
    table.column_states = [
        (col, table.columns.columns[col].header, col in table.columns_visible)
        for col in table.columns_optional
    ]


def breakpoints(table):
    bps = (
        list(table.Meta.responsive.keys()) if hasattr(table.Meta, "responsive") else []
    )
    return {"breakpoints": bps}


def save_per_page(request: HttpRequest, value: str):
    key = f"per_page:{request.resolver_match.view_name}"
    request.session[key] = value


def load_per_page(request: HttpRequest):
    key = f"per_page:{request.resolver_match.view_name}"
    return request.session[key] if key in request.session else 0


DEFAULT_APP = "django_tableaux"
DEFAULT_LIBRARY = "bootstrap4"


def get_template_library():
    if hasattr(settings, "DJANGO_TABLEAUX_LIBRARY"):
        return settings.DJANGO_TABLEAUX_LIBRARY
    return DEFAULT_LIBRARY


def get_template_prefix():
    # used for components that directly specify their template
    library = get_template_library()
    return library if library[:10] == "templates/" else f"{DEFAULT_APP}/{library}"


def build_templates_dictionary(library=""):
    """
    Returns a dictionary with key=template name (without .html suffix) and value = template path
    """
    result = {}
    library = library or get_template_library()
    app_path = apps.get_app_config(DEFAULT_APP).path
    lib_files = []
    prefix = ""
    if library != DEFAULT_LIBRARY:
        if library[:10] == "templates/":
            full_path = settings.BASE_DIR / library
            if not exists(full_path):
                raise ImproperlyConfigured(
                    f"Template library '{library}' does not exist."
                )
            prefix = full_path
        else:
            # check if in own app
            full_path = join(app_path, "templates", DEFAULT_APP, library)
            if exists(full_path):
                prefix = f"{DEFAULT_APP}/{library}"
                lib_files = [
                    f for f in listdir(full_path) if isfile(join(full_path, f))
                ]
    # iterate default library and add entries to result unless file exists in specified library
    for file in listdir(join(app_path, "templates", DEFAULT_APP, DEFAULT_LIBRARY)):
        name = file.split(".")[0]
        if file in lib_files:
            result[name] = join(prefix, file)
        else:
            result[name] = join(DEFAULT_APP, DEFAULT_LIBRARY, file)
    return result



# Generated by Django 5.0.6 on 2025-01-18 16:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('myapp', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestModel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('a', models.CharField(max_length=20)),
                ('b', models.CharField(max_length=20)),
                ('c', models.CharField(max_length=20)),
                ('d', models.CharField(max_length=20)),
            ],
        ),
    ]
